@@ -6,16 +6,25 @@ from matplotlib.lines import Line2D
 color_map = {
     "Electricity": "yellow",
     "UntreatedSewage": "saddlebrown",
+    "PrimaryEffluent": "saddlebrown",
+    "SecondaryEffluent": "saddlebrown",
+    "TertiaryEffluent": "saddlebrown",
     "TreatedSewage": "green",
     "WasteActivatedSludge": "black",
     "PrimarySludge": "black",
+    "TWAS": "black",
+    "TPS": "black",
     "Scum": "black",
     "SludgeBlend": "black",
+    "ThickenedSludgeBlend": "black",
     "Biogas": "red",
     "GasBlend": "red",
     "NaturalGas": "gray",
     "Seawater": "aqua",
     "Brine": "aqua",
+    "SurfaceWater": "cornflowerblue",
+    "Groundwater": "cornflowerblue",
+    "Stormwater": "cornflowerblue",
     "NonpotableReuse": "purple",
     "DrinkingWater": "blue",
     "PotableReuse": "blue",
@@ -63,6 +72,7 @@ def draw_graph(network, pyvis=False):
         "yellow",
         "red",
         "gray",
+        "cornflowerblue",
         "aqua",
         "purple",
         "blue",
@@ -75,6 +85,7 @@ def draw_graph(network, pyvis=False):
         "Electricity",
         "Biogas",
         "Natural Gas",
+        "Freshwater",
         "Saline Water",
         "Non-potable Reuse",
         "Drinking Water",
@@ -89,6 +100,7 @@ def draw_graph(network, pyvis=False):
         "black",
         "black",
         "black",
+        "black",
         "white",
         "black",
     ]
